/*---------------------------------------------------------------------------------------------------------
 * sndrx-bootloader.c
 *
 * Copyright (c) 2011 Robert Meyer, Frank Meyer
 *
 * ATMEGA168 @ 8 MHz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *---------------------------------------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/boot.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <inttypes.h>

#define SNDRX_AS_INCLUDE    // bad trick: save program space by including sndrx.c so we can use static functions
#include "sndrxconfig.h"
#include "sndrx.h"
#include "sndrx.c"

#define USE_LED 1           // CHANGE HERE!

#define TRUE    1
#define FALSE   0

#if USE_LED == 1

/*-----------------------------------------------------------------------------------------------------------------------
 * Status LED, change here:
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define LED_PORT        PORTD
#define LED_DDR         DDRD
#define LED             6

#define LED_ON          LED_PORT |= 1<<LED;
#define LED_OFF         LED_PORT &= ~(1<<LED);
#define LED_INIT        led_init ();

/*-----------------------------------------------------------------------------------------------------------------------
 * Initialize LED port
 *-----------------------------------------------------------------------------------------------------------------------
 */
static void
led_init (void)
{
    LED_OFF
    LED_DDR |= (1<<LED);
}

#else

#define LED_ON
#define LED_OFF
#define LED_INIT

#endif

/*-----------------------------------------------------------------------------------------------------------------------
 * Flash: erase page
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define boot_program_page_erase(pageaddr)           \
{                                                   \
    eeprom_busy_wait ();                            \
    cli ();                                         \
    boot_page_erase ((uint32_t) (pageaddr));        \
    sei ();                                         \
    boot_spm_busy_wait ();                          \
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Flash: fill page word by word
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define boot_program_page_fill(byteaddr, word)      \
{                                                   \
    cli ();                                         \
    boot_page_fill ((uint32_t) (byteaddr), word);   \
    sei ();                                         \
}

/*-----------------------------------------------------------------------------------------------------------------------
 * Flash: write page
 *-----------------------------------------------------------------------------------------------------------------------
 */
#define boot_program_page_write(pageaddr)           \
{                                                   \
    cli ();                                         \
    boot_page_write ((uint32_t) (pageaddr));        \
    sei ();                                         \
    boot_spm_busy_wait ();                          \
    boot_rww_enable ();                             \
}


static uint8_t
boot_update_flash (void)
{
    uint8_t         page_buf[2];
    uint16_t        start_addr;
    uint8_t         start_addr_h;
    uint8_t         start_addr_l;
    uint16_t        pgcnt = 0;
    uint8_t         n_pages_h;
    uint8_t         n_pages_l;
    uint16_t        n_pages;
    uint8_t         idx;
    uint8_t         chksum;
    uint8_t         ch;
    int8_t          rtc;
    uint16_t        w;

    LED_OFF

    if (sndrx_poll (&start_addr_h, 20) <= 0 ||
        sndrx_poll (&start_addr_l, 20) <= 0 ||
        sndrx_poll (&n_pages_h,    20) <= 0 ||
        sndrx_poll (&n_pages_l,    20) <= 0)
    {
        return FALSE;
    }

    n_pages    = (((uint16_t) n_pages_h)    << 8) | n_pages_l;
    start_addr = (((uint16_t) start_addr_h) << 8) | start_addr_l;

    while (pgcnt < n_pages)
    {
        LED_OFF

        chksum = 0;

        for (idx = 0; idx < SPM_PAGESIZE; )
        {
            rtc = sndrx_poll (&ch, 20);

            if (rtc <= 0)
            {
                return FALSE;
            }

            page_buf[idx & 0x01] = ch;

            chksum ^= ch;

            if (idx == 0)
            {
                boot_program_page_erase (start_addr);                           // erase at beginning of page, but after 1 char read
            }

            if (idx & 0x01)                                                     // odd?
            {
                w = page_buf[0] | (page_buf[1] << 8);
                boot_program_page_fill ((uint32_t) (start_addr + idx - 1), w);
            }
            idx++;
        }

        LED_ON

        rtc = sndrx_poll (&ch, 20);

        if (rtc <= 0 || ch != chksum)
        {
            return FALSE;
        }

        if (idx == 0)                                                           // we are ready
        {
            break;
        }
        else if (idx < SPM_PAGESIZE)                                            // error: page not completed
        {
            return FALSE;
        }

        pgcnt++;
        boot_program_page_write ((uint32_t) start_addr);        

        start_addr += SPM_PAGESIZE;
    }

    return TRUE;
}

/*-----------------------------------------------------------------------------------------------------------------------
 * main function
 *-----------------------------------------------------------------------------------------------------------------------
 */
int main (void) __attribute__ ((OS_main));                                  // saves 22 bytes

int
main (void)
{
    uint8_t     ch;
    uint8_t     temp;
    uint8_t     success = TRUE;

    sndrx_init ();                                                          // initialize SOUNDRX
    LED_INIT                                                                // initialize LED port

    temp = MCUCR;                                                           // set interrupt vectors to bootloader section
    MCUCR = temp | (1<<IVCE);
    MCUCR = temp | (1<<IVSEL);
    sei ();                                                                 // enable interrupts

    while (1)
    {
        LED_ON                                                              // switch LED on

        if (sndrx_poll (&ch, 3000) == 1 && ch == '$')                       // wait 3 seconds for the dollar character...
        {
            success = boot_update_flash ();
            _delay_ms(1000);
        }

        ringbuf_max = 0;

        if (! success)                                                      // boot_update_flash() failed...
        {
            sndrx_stop ();                                                  // stop transmission
            continue;                                                       // try again
        }

        LED_OFF                                                             // switch LED off again
        cli ();                                                             // disable interrupts
        sndrx_done ();                                                      // deactivate timer and pcint in SOUNDRX
        temp = MCUCR;                                                       // reset interrupt vectors (set to application section)
        MCUCR = temp | (1<<IVCE);
        MCUCR = temp & ~(1<<IVSEL);
        asm volatile("jmp 0x0000");                                         // jump to application
    }
}


uint8_t mcusr_mirror __attribute__ ((section (".noinit")));

void get_mcusr(void) __attribute__((naked)) __attribute__((section(".init3")));

void
get_mcusr(void)
{
    mcusr_mirror = MCUSR;
    MCUSR = 0;
    wdt_disable();
}
